/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

package org.dataone.tidy.concurrent;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.dataone.configuration.Settings;
import org.dataone.service.types.v1.Identifier;
import org.dataone.tidy.dao.ChangeRecordRepository;
import org.dataone.tidy.dao.MergeResultRepository;
import org.dataone.tidy.dao.SystemMetadataTidyDao;
import org.dataone.tidy.merge.MNSystemMetadataDelegate;

/**
 * 
 * Provide an interface to retrieve a Job based on a pid
 * This interface will allow for testing of the MergeExecutorService
 * 
 * @author waltz
 */
public class TidyJobFactory {

    private static final Log logger = LogFactory.getLog(TidyJobFactory.class.getName());
    private static String tinyJobClassName = Settings.getConfiguration().getString(
            "tidy.job.class.name", "org.dataone.tidy.merge.MergeJob");

    private SystemMetadataTidyDao systemMetadataTidyDao;
    private MergeResultRepository mergeResultRepo;
    private ChangeRecordRepository changeRecordRepo;
    private MNSystemMetadataDelegate mnReadClient;

    private TidyJobFactory() {
    }

    public TidyJobFactory(SystemMetadataTidyDao systemMetadataTidyDao,
            MergeResultRepository mergeResultRepo, ChangeRecordRepository changeRecordRepo,
            MNSystemMetadataDelegate readClient) {
        this.systemMetadataTidyDao = systemMetadataTidyDao;
        this.mergeResultRepo = mergeResultRepo;
        this.changeRecordRepo = changeRecordRepo;
        this.mnReadClient = readClient;
    }

    public TidyJob getTidyJob(Identifier identifier) {

        TidyJob tidyJob = null;
        try {
            Class tidyJobClass;
            tidyJobClass = Class.forName(getTinyJobClassName());

            tidyJob = (TidyJob) tidyJobClass.newInstance();
            tidyJob.setPid(identifier);
            tidyJob.setSystemMetadataTidyDao(systemMetadataTidyDao);
            tidyJob.setMergeResultRepository(mergeResultRepo);
            tidyJob.setChangeRecordRepository(changeRecordRepo);
            tidyJob.setMnReadClient(mnReadClient);

        } catch (InstantiationException ex) {
            logger.error(ex);
        } catch (IllegalAccessException ex) {
            logger.error(ex);
        } catch (ClassNotFoundException ex) {
            logger.error(ex);
        } finally {
            if (tidyJob == null) {
                throw new NullPointerException(getTinyJobClassName() + " could not be instantiated");
            }
        }

        return tidyJob;
    }

    public static String getTinyJobClassName() {
        return tinyJobClassName;
    }

    public static void setTinyJobClassName(String tinyJobClassName) {
        TidyJobFactory.tinyJobClassName = tinyJobClassName;
    }
}
