/**
 * This work was created by participants in the DataONE project, and is
 * jointly copyrighted by participating institutions in DataONE. For 
 * more information on DataONE, see our web site at http://dataone.org.
 *
 *   Copyright ${year}
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 * 
 * $Id$
 */
package org.dataone.tidy.merge.strategy;

import java.io.IOException;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.apache.commons.lang.SerializationUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.log4j.Appender;

import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.spi.ErrorHandler;
import org.apache.log4j.spi.Filter;
import org.apache.log4j.spi.Layout;
import org.apache.log4j.spi.LoggingEvent;
import org.dataone.service.exceptions.NotFound;
import org.dataone.service.exceptions.ServiceFailure;
import org.dataone.service.types.TypeCompareUtil;
import org.dataone.service.types.v1.AccessPolicy;
import org.dataone.service.types.v1.AccessRule;
import org.dataone.service.types.v1.Identifier;
import org.dataone.service.types.v1.NodeReference;
import org.dataone.service.types.v1.Permission;
import org.dataone.service.types.v1.Replica;
import org.dataone.service.types.v1.ReplicationPolicy;
import org.dataone.service.types.v1.ReplicationStatus;
import org.dataone.service.types.v1.Subject;
import org.dataone.service.types.v1.SystemMetadata;
import org.dataone.tidy.dao.ChangeRecord;
import org.dataone.tidy.merge.MNSystemMetadataDelegate;
import org.dataone.tidy.merge.MergeReport;
import org.dataone.tidy.merge.strategy.MergeSystemMetadata;
import org.jibx.runtime.JiBXException;
import org.junit.Assert;
import org.junit.BeforeClass;
import org.junit.Test;

//TODO: test getting the AuthMN sysmeta method (AcceptAuthMNAccessPolicy)
public class MergeSystemMetadataTest  {

	
    private static Logger rootLogger = Logger.getRootLogger();
    private static Log logger = LogFactory.getLog(MergeSystemMetadataTest.class);
    protected static SystemMetadata minimalSysmeta = null;
    
    // for testing ascendent sysmeta
    protected static SystemMetadata sysmeta_v4_t12_rh0;
    protected static SystemMetadata sysmeta_v4_t13_rh1;
    protected static SystemMetadata sysmeta_v5_t11_rh2;
    protected static SystemMetadata sysmeta_v5_t11_rh3;
    protected static SystemMetadata sysmeta_v5_t10_rh4;
    protected static SystemMetadata sysmeta_v5_t10_rh5;
    
    @BeforeClass
    public static void before() {
    	minimalSysmeta = new SystemMetadata();
    	minimalSysmeta.setIdentifier(new Identifier());
    	minimalSysmeta.getIdentifier().setValue("foo");
    	minimalSysmeta.setAuthoritativeMemberNode(new NodeReference());
    	minimalSysmeta.getAuthoritativeMemberNode().setValue("mnode");
    	

    	
    	sysmeta_v4_t12_rh0 = clone(minimalSysmeta);
    	sysmeta_v4_t12_rh0.setRightsHolder(new Subject());
    	sysmeta_v4_t12_rh0.getRightsHolder().setValue("rh_0");
    	sysmeta_v4_t12_rh0.setSerialVersion(new BigInteger("4"));
    	sysmeta_v4_t12_rh0.setDateSysMetadataModified(new Date(1231230012));
    	
    	sysmeta_v4_t13_rh1 = clone(minimalSysmeta);
    	sysmeta_v4_t13_rh1.setRightsHolder(new Subject());
    	sysmeta_v4_t13_rh1.getRightsHolder().setValue("rh_1");
    	sysmeta_v4_t13_rh1.setSerialVersion(new BigInteger("4"));
    	sysmeta_v4_t13_rh1.setDateSysMetadataModified(new Date(1231230013));
    	
    	sysmeta_v5_t11_rh2 = clone(minimalSysmeta);
    	sysmeta_v5_t11_rh2.setRightsHolder(new Subject());
    	sysmeta_v5_t11_rh2.getRightsHolder().setValue("rh_2");
    	sysmeta_v5_t11_rh2.setSerialVersion(new BigInteger("5"));
    	sysmeta_v5_t11_rh2.setDateSysMetadataModified(new Date(1231230011));
    	
    	sysmeta_v5_t11_rh3 = clone(minimalSysmeta);
    	sysmeta_v5_t11_rh3.setRightsHolder(new Subject());
    	sysmeta_v5_t11_rh3.getRightsHolder().setValue("rh_3");
    	sysmeta_v5_t11_rh3.setSerialVersion(new BigInteger("5"));
    	sysmeta_v5_t11_rh3.setDateSysMetadataModified(new Date(1231230011));
    	
    	sysmeta_v5_t10_rh4 = clone(minimalSysmeta);
    	sysmeta_v5_t10_rh4.setRightsHolder(new Subject());
    	sysmeta_v5_t10_rh4.getRightsHolder().setValue("rh_4");
    	sysmeta_v5_t10_rh4.setSerialVersion(new BigInteger("5"));
    	sysmeta_v5_t10_rh4.setDateSysMetadataModified(new Date(1231230010));
    	
    	sysmeta_v5_t10_rh5 = clone(minimalSysmeta);
    	sysmeta_v5_t10_rh5.setRightsHolder(new Subject());
    	sysmeta_v5_t10_rh5.getRightsHolder().setValue("rh_5");
    	sysmeta_v5_t10_rh5.setSerialVersion(new BigInteger("5"));
    	sysmeta_v5_t10_rh5.setDateSysMetadataModified(new Date(1231230010));
    	
    	
    }
    @Test
    public void testTrue() {
    }
    
    
    /*
     * The test harness for testing ascendent sysmetas
     */
    private void determineAscendentSysmetaAssay(Map<String,SystemMetadata> sysmetas, String expected)
            throws IOException, InstantiationException, IllegalAccessException, JiBXException {

    	MergeReport report = new MergeReport(new Identifier());
    	report.getIdentifier().setValue("foo");
    	
    	
    	
        MergeSystemMetadata msm = new MergeSystemMetadata(report);
        String actual = msm.determineAscendentSystemMetadata(sysmetas);

        Assert.assertTrue(actual.equals(expected));
    }
    

    @Test
    public void testDetermineAscendentSysMeta_OneLatestSerialVersion() throws Exception 
    {   	
    	Map<String,SystemMetadata> sysmeta = new LinkedHashMap<String,SystemMetadata>();
    	sysmeta.put("first", clone(sysmeta_v4_t12_rh0));
    	sysmeta.put("second", clone(sysmeta_v4_t13_rh1));
    	sysmeta.put("third", clone(sysmeta_v5_t11_rh2));
    
    	String expected = "third";
    	
    	determineAscendentSysmetaAssay(sysmeta, expected);
    }

    @Test
    public void testDetermineAscendentSysMeta_TwoLatestSerialVersions_SameDate() throws Exception {

        
        Map<String, SystemMetadata> sysmeta = new LinkedHashMap<String,SystemMetadata>();
        sysmeta.put("first", clone(sysmeta_v4_t13_rh1));
        sysmeta.put("second", clone(sysmeta_v5_t11_rh2));
        sysmeta.put("third", clone(sysmeta_v5_t11_rh3));
    
    	String expected = "second";
    	
    	determineAscendentSysmetaAssay(sysmeta, expected);
    }

    @Test
    public void testDetermineAscendentSysMeta_TwoLatestSerialVersions_OneLatestDate()
            throws Exception {

        
        Map<String,SystemMetadata> sysmeta = new LinkedHashMap<String, SystemMetadata>();
    	sysmeta.put("first", clone(sysmeta_v4_t13_rh1));
    	sysmeta.put("second", clone(sysmeta_v5_t11_rh2));
    	sysmeta.put("third", clone(sysmeta_v5_t10_rh4));
    
    	String expected ="second";
    	
    	determineAscendentSysmetaAssay(sysmeta, expected);
    }

    @Test
    public void testDetermineAscendentSysMeta_AllSameSerialVersion_OneLatestDate() throws Exception {

        
        Map<String, SystemMetadata> sysmeta = new LinkedHashMap<String, SystemMetadata>();
    	sysmeta.put("first", clone(sysmeta_v5_t11_rh3));
    	sysmeta.put("second", clone(sysmeta_v5_t10_rh4));
    	sysmeta.put("third", clone(sysmeta_v5_t10_rh5));
    
    	String expected = "first";
    	
    	determineAscendentSysmetaAssay(sysmeta, expected);
    }

    
     
    
    
    @Test
    public void testMergeArchive_nonAscendentArchived() throws Exception
    {
    	Map<String,SystemMetadata> sysmeta = new LinkedHashMap<String,SystemMetadata>();
    	sysmeta.put("first", clone(minimalSysmeta));
    	
    	sysmeta.put("second", clone(minimalSysmeta));
    	sysmeta.put("third", clone(minimalSysmeta));
    	    	
    	// set archive to True for non-ascendent
    	sysmeta.get("second").setArchived(true);
   	
    	SystemMetadata expected = clone(sysmeta.get("second"));
    	
    	runTest("first", expected, sysmeta, "archived", null, null, null);
    }
    
    
    @Test
    public void testMergeArchive_ascendentArchived() throws Exception
    {
    	Map<String,SystemMetadata> sysmeta = new LinkedHashMap<String,SystemMetadata>();
    	sysmeta.put("first", clone(minimalSysmeta));
    	
    	sysmeta.put("second", clone(minimalSysmeta));
    	sysmeta.put("third", clone(minimalSysmeta));

    	    	
    	// set archive to True for ascendent
    	sysmeta.get("first").setArchived(true);

    	
    	SystemMetadata expected = clone(sysmeta.get("first"));
    	
    	runTest("first", expected, sysmeta, "archived", null, null, null);
    }
    
    
    @Test
    public void testMergeDateSysMetaModified() throws Exception
    {
    	Map<String,SystemMetadata> sysmeta = new LinkedHashMap<String,SystemMetadata>();
    	sysmeta.put("first", clone(minimalSysmeta));
    	sysmeta.get("first").setDateSysMetadataModified(new Date());
    	sysmeta.get("first").setSerialVersion(new BigInteger("3"));
    	
    	sysmeta.put("second", clone(sysmeta.get("first")));
    	sysmeta.put("third", clone(sysmeta.get("first")));
    	    	
    	// give a newer Date to one of the others
    	Thread.sleep(1000);
    	Date latestDate = new Date();
    	sysmeta.get("third").setDateSysMetadataModified(latestDate);
    	
    	SystemMetadata expected = clone(sysmeta.get("third"));
    	
    	runTest("first", expected, sysmeta, "modifiedDate", null, null, null);
    }
    
    
    
    @Test
    public void testMergeDateSysMetaModified_ascendent() throws Exception
    {
    	Map<String,SystemMetadata> sysmeta = new LinkedHashMap<String,SystemMetadata>();
    	sysmeta.put("first", clone(minimalSysmeta));
    	sysmeta.get("first").setDateSysMetadataModified(new Date());
    	
    	sysmeta.put("second", clone(sysmeta.get("first")));
    	sysmeta.put("third", clone(sysmeta.get("first")));
    	    	
    	// give a newer Date to one of the others
    	Thread.sleep(1000);
    	Date latestDate = new Date();
    	sysmeta.get("first").setDateSysMetadataModified(latestDate);
    	
    	SystemMetadata expected = clone(sysmeta.get("first"));
    	
    	runTest("first", expected, sysmeta, "modifiedDate", null, null, null);
    }
    
    
    @Test
    public void testMergeDateSysMetaModified_allSame() throws Exception
    {
    	Map<String,SystemMetadata> sysmeta = new LinkedHashMap<String,SystemMetadata>();
    	sysmeta.put("first", clone(minimalSysmeta));
    	sysmeta.get("first").setDateSysMetadataModified(new Date());
    	
    	sysmeta.put("second", clone(sysmeta.get("first")));
    	sysmeta.put("third", clone(sysmeta.get("first")));
    	
    	SystemMetadata expected = clone(sysmeta.get("first"));
    	
    	runTest("first",expected, sysmeta,"modifiedDate",null, null,null);
    }
    

 
    /**
     * the mergeTarget should take the value of the one with the obsoletedBy value set.
     * @throws Exception
     */
    @Test
    public void testMergeObsoletedBy() throws Exception
    {
    	Map<String,SystemMetadata> sysmeta = new LinkedHashMap<String,SystemMetadata>();
    	sysmeta.put("first", clone(minimalSysmeta));
    	
    	sysmeta.put("second", clone(sysmeta.get("first")));
    	sysmeta.get("second").setObsoletedBy(new Identifier());
    	sysmeta.get("second").getObsoletedBy().setValue("yyyyy");
    	
    	sysmeta.put("third", clone(minimalSysmeta));
    	
    	SystemMetadata expected = clone(sysmeta.get("second"));
    	
    	runTest("first", expected, sysmeta,
    			"obsoletedBy", null, null, null);
    }
    
    
    /**
     * Tests mainly the logging behavior when 2 different values for ObsoletedBy
     * are encountered.
     */
    @Test
    public void testMergeObsoletedBy_conflicting_values() throws Exception
    {
    	Map<String,SystemMetadata> sysmeta = new LinkedHashMap<String,SystemMetadata>();
    	sysmeta.put("first", clone(minimalSysmeta));
    	sysmeta.get("first").setObsoletedBy(new Identifier());
    	sysmeta.get("first").getObsoletedBy().setValue("xxxxx");
    	
    	sysmeta.put("second", clone(sysmeta.get("first")));
    	sysmeta.get("second").getObsoletedBy().setValue("yyyyy");
    	
    	sysmeta.put("third", clone(minimalSysmeta));
    	
    	SystemMetadata expected = clone(sysmeta.get("first"));
    	
    	runTest("third", expected, sysmeta,
    			"obsoletedBy", null, Level.ERROR, "Multiple values found for ObsoletedBy field");
    }
    
    
    
    
    /**
     * the mergeTarget should take the value of the one with the obsoletes value set.
     * @throws Exception
     */
    @Test
    public void testMergeObsoletes() throws Exception
    {
    	Map<String,SystemMetadata> sysmeta = new LinkedHashMap<String,SystemMetadata>();
    	sysmeta.put("first", clone(minimalSysmeta));
    	
    	sysmeta.put("second", clone(sysmeta.get("first")));
    	sysmeta.get("second").setObsoletes(new Identifier());
    	sysmeta.get("second").getObsoletes().setValue("yyyyy");
    	
    	sysmeta.put("third", clone(minimalSysmeta));
    	
    	SystemMetadata expected = clone(sysmeta.get("second"));
    	
    	runTest("first", expected, sysmeta,
    			"obsoletes", null, null, null);
    }
    
    
    /**
     * Tests mainly the logging behavior when 2 different values for Obsoletes 
     * are encountered.
     */
    @Test
    public void testMergeObsoletes_conflicting_values() throws Exception
    {
    	Map<String,SystemMetadata> sysmeta = new LinkedHashMap<String,SystemMetadata>();
    	sysmeta.put("first", clone(minimalSysmeta));
    	sysmeta.get("first").setObsoletes(new Identifier());
    	sysmeta.get("first").getObsoletes().setValue("xxxxx");
    	
    	sysmeta.put("second", clone(sysmeta.get("first")));
    	sysmeta.get("second").getObsoletes().setValue("yyyyy");
    	
    	sysmeta.put("third", clone(minimalSysmeta));
    	
    	SystemMetadata expected = clone(sysmeta.get("first"));
    	
    	runTest("third", expected, sysmeta,
    			"obsoletes", null, Level.ERROR, "Multiple values found for Obsoletes field");
    }
   
    @Test
    public void testMergeReplicaList() throws Exception
    {
    	Map<String,SystemMetadata> sysmeta = new LinkedHashMap<String,SystemMetadata>();
    	sysmeta.put("first", clone(minimalSysmeta));
    	sysmeta.get("first").addReplica(new Replica());
    	sysmeta.get("first").getReplica(0).setReplicaMemberNode(new NodeReference());
    	sysmeta.get("first").getReplica(0).getReplicaMemberNode().setValue("buddyNode");
    	sysmeta.get("first").getReplica(0).setReplicaVerified(new Date(123123123));
    	sysmeta.get("first").getReplica(0).setReplicationStatus(ReplicationStatus.COMPLETED);

    	sysmeta.get("first").addReplica(new Replica());
    	sysmeta.get("first").getReplica(1).setReplicaMemberNode(new NodeReference());
    	sysmeta.get("first").getReplica(1).getReplicaMemberNode().setValue("friendNode");
    	sysmeta.get("first").getReplica(1).setReplicaVerified(new Date(1231231234));
    	sysmeta.get("first").getReplica(1).setReplicationStatus(ReplicationStatus.COMPLETED);

    	
    	sysmeta.put("second", clone(sysmeta.get("first")));    	
    	// same node, same date, different state
    	sysmeta.get("second").getReplica(0).setReplicationStatus(ReplicationStatus.INVALIDATED);
    	// same node, later date, different status (lower)
    	sysmeta.get("second").getReplica(1).setReplicaVerified(new Date(1231231239));
    	sysmeta.get("second").getReplica(1).setReplicationStatus(ReplicationStatus.REQUESTED);

    	// different first replica
    	sysmeta.put("third", clone(sysmeta.get("first")));
    	sysmeta.get("third").getReplica(0).getReplicaMemberNode().setValue("anotherNode");
    	sysmeta.get("third").getReplica(0).setReplicaVerified(new Date(1231231230));
    	sysmeta.get("third").getReplica(0).setReplicationStatus(ReplicationStatus.COMPLETED);
    	
    	
    	SystemMetadata expected = clone(sysmeta.get("first"));
    	expected.getReplica(0).setReplicationStatus(ReplicationStatus.INVALIDATED);
    	expected.getReplica(1).setReplicaVerified(new Date(1231231239));
       	expected.getReplica(1).setReplicationStatus(ReplicationStatus.REQUESTED);
       	expected.addReplica( (Replica) SerializationUtils.clone(sysmeta.get("third").getReplica(0)));

       	runTest("first", expected, sysmeta, "replicaList", null, null, null);
    	
    }
    
    /**
     * Ascendent rightsholder always wins
     * @throws Exception 
     */
//    @Test
    public void testMergeRightsHolder() throws Exception
    {
    	Map<String,SystemMetadata> sysmeta = new LinkedHashMap<String,SystemMetadata>();
    	sysmeta.put("first", clone(minimalSysmeta));
    	sysmeta.get("first").setRightsHolder(new Subject());
    	sysmeta.get("first").getRightsHolder().setValue("me");
    	
    	sysmeta.put("second", clone(sysmeta.get("first")));
    	sysmeta.get("second").getRightsHolder().setValue("she");
    	
    	sysmeta.put("third", clone(sysmeta.get("first")));
    	sysmeta.get("third").getRightsHolder().setValue("he");
    	
    	SystemMetadata expected = clone(sysmeta.get("first"));
    	runTest("first", expected, sysmeta, "rightsHolder", null, Level.WARN, "Rightsholder value");
    }
    
    /**
     * test that isReplicationAllows is set to true if any are already true,
     * and if so, that the number of replicas is the max of the individual ones.
     * @throws Exception 
     */
    @Test
    public void testMergeReplicationPolicyInfo() throws Exception
    {
    	Map<String,SystemMetadata> sysmeta = new LinkedHashMap<String,SystemMetadata>();
    	sysmeta.put("first", clone(minimalSysmeta));

    	sysmeta.put("second", clone(sysmeta.get("first")));
    	sysmeta.get("second").setReplicationPolicy(new ReplicationPolicy());
    	sysmeta.get("second").getReplicationPolicy().setReplicationAllowed(true);
    	sysmeta.get("second").getReplicationPolicy().setNumberReplicas(3);
    	
    	sysmeta.put("third", clone(sysmeta.get("second")));
    	sysmeta.get("third").getReplicationPolicy().setNumberReplicas(2);
    	
    	SystemMetadata expected = clone(sysmeta.get("second"));
    	
    	runTest("first", expected, sysmeta,
    			"replicaPolicyInfo", null, null, null);
    }
    
    
    
    /**
     * test that isReplicationAllows is set to false if all are false (or null)
     * and that the number of replicas = 0 when allowed is false.
     * @throws Exception 
     */
    @Test
    public void testMergeReplicationPolicyInfo_NotAllowed() throws Exception
    {
    	Map<String,SystemMetadata> sysmeta = new LinkedHashMap<String,SystemMetadata>();
    	sysmeta.put("first", clone(minimalSysmeta));

    	sysmeta.put("second", clone(sysmeta.get("first")));
    	sysmeta.get("second").setReplicationPolicy(new ReplicationPolicy());
    	sysmeta.get("second").getReplicationPolicy().setReplicationAllowed(false);
    	sysmeta.get("second").getReplicationPolicy().setNumberReplicas(2);
    	
    	sysmeta.put("third", clone(sysmeta.get("second")));
    	
    	SystemMetadata expected = clone(sysmeta.get("first"));
    	
    	runTest("first", expected, sysmeta,
    			"replicaPolicyInfo", null, null, null);
    }
    
    
    /**
     * The Preferred lists are ordered, and so when merging, the resulting superset
     * needs to maintain the original ordering, and this is done via collating the
     * lists, as in: [a1,a2,a3...], [b1,b2...], [c1,c2,..] => [a1,b1,c1,a2,b2,c2,...]
     * The order of lists prioritizes the list of the ascendent (target) sysmeta.
     * 
     * Will create 3 lists all with different values to test the collation  
     * @throws Exception 
     */
    @Test 
    public void testMergePreferredNodes_testSimpleCollation() throws Exception
    {
    	Map<String,SystemMetadata> sysmeta = new LinkedHashMap<String,SystemMetadata>();

    	sysmeta.put("first", clone(minimalSysmeta));
    	
    	sysmeta.get("first").setReplicationPolicy(new ReplicationPolicy());
    	sysmeta.get("first").getReplicationPolicy().setPreferredMemberNodeList(
    			buildNodeReferenceList( new String[]{"a1","a2","a3","a4"}));
    	
    	sysmeta.put("second", clone(minimalSysmeta));
    	
    	sysmeta.get("second").setReplicationPolicy(new ReplicationPolicy());
    	sysmeta.get("second").getReplicationPolicy().setPreferredMemberNodeList(
    			buildNodeReferenceList( new String[]{"b1","b2","b3","b4"}));
    	
    	sysmeta.put("third", clone(minimalSysmeta));
    	
    	sysmeta.get("third").setReplicationPolicy(new ReplicationPolicy());
    	sysmeta.get("third").getReplicationPolicy().setPreferredMemberNodeList(
    			buildNodeReferenceList( new String[]{"c1","c2"}));
    	
    	SystemMetadata expected = clone(minimalSysmeta);
    	expected.setReplicationPolicy(new ReplicationPolicy());
    	expected.getReplicationPolicy().setPreferredMemberNodeList(
    			buildNodeReferenceList( new String[]{"c1","a1","b1","c2","a2","b2","a3","b3","a4","b4"}));
 
    	runTest("third",expected, sysmeta,
    			"preferred", null, null, "Preferred will add");
    }
    
    /** 
    * Will create 3 lists overlapping values, and different list lengths  
    * @throws Exception 
    */
    @Test 
    public void testMergePreferredNodes_with_duplicates() throws Exception
    {
    	Map<String,SystemMetadata> sysmeta = new LinkedHashMap<String,SystemMetadata>();

    	sysmeta.put("first", clone(minimalSysmeta));

    	sysmeta.get("first").setReplicationPolicy(new ReplicationPolicy());
    	sysmeta.get("first").getReplicationPolicy().setPreferredMemberNodeList(
    			buildNodeReferenceList( new String[]{"a","b","c","f","h"}));

    	sysmeta.put("second", clone(minimalSysmeta));

    	sysmeta.get("second").setReplicationPolicy(new ReplicationPolicy());
    	sysmeta.get("second").getReplicationPolicy().setPreferredMemberNodeList(
    			buildNodeReferenceList( new String[]{"a","b","d","g","i","j"}));

    	sysmeta.put("third", clone(minimalSysmeta));

    	sysmeta.get("third").setReplicationPolicy(new ReplicationPolicy());
    	sysmeta.get("third").getReplicationPolicy().setPreferredMemberNodeList(
    			buildNodeReferenceList( new String[]{"a","b","e"}));

    	SystemMetadata expected = clone(minimalSysmeta);
    	expected.setReplicationPolicy(new ReplicationPolicy());
    	expected.getReplicationPolicy().setPreferredMemberNodeList(
    			buildNodeReferenceList( new String[]{"a","b","c","d","e","f","g","h","i","j"}));

    	runTest("first",expected, sysmeta,
    			"preferred", null, null, "Preferred will add");
    }
    
  
    /**
     * the blocked nodes should get merged, (and sorted). Additions from non-ascendents
     * should be logged. 
     * @throws Exception 
     */
    @Test
    public void testMergeBlockedNodes() throws Exception
    {
    	Map<String,SystemMetadata> sysmeta = new LinkedHashMap<String,SystemMetadata>();
    	
    	sysmeta.put("first", clone(minimalSysmeta));
    	sysmeta.get("first").setReplicationPolicy(new ReplicationPolicy());
    	sysmeta.get("first").getReplicationPolicy().setBlockedMemberNodeList(
    			buildNodeReferenceList( new String[]{"a","b","c","g"}));
    	
    	sysmeta.put("second", clone(minimalSysmeta));
    	sysmeta.get("second").setReplicationPolicy(new ReplicationPolicy());
    	sysmeta.get("second").getReplicationPolicy().setBlockedMemberNodeList(
    			buildNodeReferenceList( new String[]{"a","b","d"}));
    	
    	sysmeta.put("third", clone(minimalSysmeta));
    	sysmeta.get("third").setReplicationPolicy(new ReplicationPolicy());
    	sysmeta.get("third").getReplicationPolicy().setBlockedMemberNodeList(
    			buildNodeReferenceList( new String[]{"c","d","e","f"}));

    	SystemMetadata expected = clone(minimalSysmeta);
    	expected.setReplicationPolicy(new ReplicationPolicy());
    	expected.getReplicationPolicy().setBlockedMemberNodeList(
    			buildNodeReferenceList( new String[]{"a","b","c","d","e","f","g"}));
    	
    	runTest("first", expected, sysmeta,
    			"blocked", null, null, "BlockedNodeList will add");	
    }
    
    private List<NodeReference> buildNodeReferenceList(String[] stringValues) {
    	List<NodeReference> refList = new ArrayList<NodeReference>();
    	
    	for (String v : stringValues) {
    		NodeReference nr = new NodeReference();
    		nr.setValue(v);
    		refList.add(nr);
    	}
    	return refList;
    }
    
    
    /**
     * The mergeTarget wins every time, and discrepancies logged.
     * @throws Exception 
     */
    @Test 
    public void testMergAccessPolicy() throws Exception
    {
    	Map<String,SystemMetadata> sysmeta = new LinkedHashMap<String,SystemMetadata>();
    	sysmeta.put("first", clone(minimalSysmeta));
    	
    	sysmeta.get("first").setAccessPolicy(new AccessPolicy());
    	sysmeta.get("first").getAccessPolicy().addAllow(new AccessRule());
    	sysmeta.get("first").getAccessPolicy().getAllow(0).addPermission(Permission.WRITE);
    	sysmeta.get("first").getAccessPolicy().getAllow(0).addSubject(new Subject());
    	sysmeta.get("first").getAccessPolicy().getAllow(0).getSubject(0).setValue("jpMorgan");
    	
    	// vary the permission
    	sysmeta.put("second", clone(sysmeta.get("first")));
    	sysmeta.get("second").getAccessPolicy().clearAllowList();
    	sysmeta.get("second").getAccessPolicy().addAllow(new AccessRule());
    	sysmeta.get("second").getAccessPolicy().getAllow(0).addPermission(Permission.READ);
    	sysmeta.get("second").getAccessPolicy().getAllow(0).addSubject(new Subject());
    	sysmeta.get("second").getAccessPolicy().getAllow(0).getSubject(0).setValue("jpMorgan");

    	// vary the subject
    	sysmeta.put("third", clone(sysmeta.get("first")));
    	sysmeta.get("third").getAccessPolicy().getAllow(0).getSubject(0).setValue("LehmanBros");
    	
    	SystemMetadata expected = clone(sysmeta.get("first"));
    	
    	ConstantAuthMNSysmetaProxy prx = new ConstantAuthMNSysmetaProxy();
    	SystemMetadata temp = prx.getSystemMetadata(new NodeReference(), new Identifier());
    	expected.setAccessPolicy(temp.getAccessPolicy());
    	
    	MergeReport report = runTest("first", expected, sysmeta,
    			"mnAccessPolicy", null, null, "");	
    	
    	Assert.assertTrue(report.getAuthMnSystemMetadata() != null);
    	Assert.assertTrue(report.getProcessingException() == null);
    }
 
    
    
    @Test 
    public void testMergAccessPolicy_NotFound() throws Exception
    {
    	Map<String,SystemMetadata> sysmeta = new LinkedHashMap<String,SystemMetadata>();
    	sysmeta.put("first", clone(minimalSysmeta));
    	
    	sysmeta.get("first").setAccessPolicy(new AccessPolicy());
    	sysmeta.get("first").getAccessPolicy().addAllow(new AccessRule());
    	sysmeta.get("first").getAccessPolicy().getAllow(0).addPermission(Permission.WRITE);
    	sysmeta.get("first").getAccessPolicy().getAllow(0).addSubject(new Subject());
    	sysmeta.get("first").getAccessPolicy().getAllow(0).getSubject(0).setValue("jpMorgan");
    	
    	// vary the permission
    	sysmeta.put("second", clone(sysmeta.get("first")));
    	sysmeta.get("second").getAccessPolicy().clearAllowList();
    	sysmeta.get("second").getAccessPolicy().addAllow(new AccessRule());
    	sysmeta.get("second").getAccessPolicy().getAllow(0).addPermission(Permission.READ);
    	sysmeta.get("second").getAccessPolicy().getAllow(0).addSubject(new Subject());
    	sysmeta.get("second").getAccessPolicy().getAllow(0).getSubject(0).setValue("jpMorgan");

    	// vary the subject
    	sysmeta.put("third", clone(sysmeta.get("first")));
    	sysmeta.get("third").getAccessPolicy().getAllow(0).getSubject(0).setValue("LehmanBros");
    	
    	SystemMetadata expected = clone(sysmeta.get("first"));
    	
    	MergeReport report = runTest("first", expected, sysmeta,
    			"mnAccessPolicy", new ExceptionAuthMNSysmetaProxy(NotFound.class), null, "");	
    	
    	Assert.assertTrue(report.getAuthMnSystemMetadata() == null);
    	Assert.assertTrue(report.getProcessingException() == null);
    }
    

    
    @Test 
    public void testMergAccessPolicy_ServiceFailure() throws Exception
    {
    	Map<String,SystemMetadata> sysmeta = new LinkedHashMap<String,SystemMetadata>();
    	sysmeta.put("first", clone(minimalSysmeta));
    	
    	sysmeta.get("first").setAccessPolicy(new AccessPolicy());
    	sysmeta.get("first").getAccessPolicy().addAllow(new AccessRule());
    	sysmeta.get("first").getAccessPolicy().getAllow(0).addPermission(Permission.WRITE);
    	sysmeta.get("first").getAccessPolicy().getAllow(0).addSubject(new Subject());
    	sysmeta.get("first").getAccessPolicy().getAllow(0).getSubject(0).setValue("jpMorgan");
    	
    	// vary the permission
    	sysmeta.put("second", clone(sysmeta.get("first")));
    	sysmeta.get("second").getAccessPolicy().clearAllowList();
    	sysmeta.get("second").getAccessPolicy().addAllow(new AccessRule());
    	sysmeta.get("second").getAccessPolicy().getAllow(0).addPermission(Permission.READ);
    	sysmeta.get("second").getAccessPolicy().getAllow(0).addSubject(new Subject());
    	sysmeta.get("second").getAccessPolicy().getAllow(0).getSubject(0).setValue("jpMorgan");

    	// vary the subject
    	sysmeta.put("third", clone(sysmeta.get("first")));
    	sysmeta.get("third").getAccessPolicy().getAllow(0).getSubject(0).setValue("LehmanBros");
    	
    	SystemMetadata expected = clone(sysmeta.get("first"));
 
    	
    	MergeReport report = runTest("first", expected, sysmeta,
    			"mnAccessPolicy", new ExceptionAuthMNSysmetaProxy(ServiceFailure.class), null, "");	

    	Assert.assertTrue(report.getAuthMnSystemMetadata() == null);
    	Assert.assertTrue(report.getProcessingException() instanceof ServiceFailure);
    	
    }

    
    
    @Test 
    public void testMergeAuthoritativeMemberNode() throws Exception 
    {
    	Map<String,SystemMetadata> sysmeta = new LinkedHashMap<String,SystemMetadata>();
    	sysmeta.put("first", clone(minimalSysmeta));
    	sysmeta.get("first").setAuthoritativeMemberNode(new NodeReference());
    	sysmeta.get("first").getAuthoritativeMemberNode().setValue("mine");
    	
    	sysmeta.put("second", clone(sysmeta.get("first")));
    	sysmeta.get("second").getAuthoritativeMemberNode().setValue("his");
    	
    	sysmeta.put("third", clone(sysmeta.get("first")));
    	sysmeta.get("third").getAuthoritativeMemberNode().setValue("hers");
    	
    	SystemMetadata expected = clone(sysmeta.get("first"));
    	runTest("first", expected, sysmeta, "authMN", null, Level.WARN, "AuthoritativeMemberNode value");
    }
    
    
    
    /*
     * workhorse method to run merge tests
     */
    private MergeReport runTest(String mergeTarget, SystemMetadata expected,
    		Map<String,SystemMetadata> sysMetas,
    		String mergeMethod, MNSystemMetadataDelegate mnFetcher,
    		Level level, String messageContains) throws Exception 
    {
    	
    	if (mnFetcher == null) {
    		mnFetcher = new ConstantAuthMNSysmetaProxy();
    	}


    	MergeReport report = new MergeReport(new Identifier());
    	report.getIdentifier().setValue("foo");

    	// run the method under test
    	MergeSystemMetadata msm = new MergeSystemMetadata(report, mnFetcher);
    	boolean sort = true;

    	if (mergeMethod.equalsIgnoreCase("archived")) {
    		msm.applyArchivedIfAnyArchived(mergeTarget, sysMetas);
    	} else if (mergeMethod.equalsIgnoreCase("authMN")) {
    		msm.mergeAuthoritativeMemberNode(mergeTarget, sysMetas);
    	} else if (mergeMethod.equalsIgnoreCase("accessPolicy")) {
    		msm.mergeAccessPolicy(mergeTarget, sysMetas);
    	} else if (mergeMethod.equalsIgnoreCase("mnAccessPolicy")) {
    		try {
    			msm.acceptAuthMNAccessPolicy(mergeTarget, sysMetas);
    		} catch (Exception e) {
    			report.setProcessingException(e);
    		}
    	} else if (mergeMethod.equalsIgnoreCase("blocked")) {
    		msm.mergeBlocked(mergeTarget, sysMetas);
    	} else if (mergeMethod.equalsIgnoreCase("modifiedDate")) {
    		msm.applyLatestDateSystemMetadataModified(mergeTarget, sysMetas);
    	} else if (mergeMethod.equalsIgnoreCase("obsoletedBy")) {
    		msm.useAscendentsObsoletedBy(mergeTarget, sysMetas);
    	} else if (mergeMethod.equalsIgnoreCase("obsoletes")) {
    		msm.useAscendentsObsoletes(mergeTarget, sysMetas);
    	} else if (mergeMethod.equalsIgnoreCase("preferred")) {
    		sort = false;
    		msm.collateAllPreferredLists(mergeTarget, sysMetas);
    	} else if (mergeMethod.equalsIgnoreCase("replicaList")) {
    		msm.mergeReplicaList(mergeTarget, sysMetas);
    	} else if (mergeMethod.equalsIgnoreCase("replicaPolicyInfo")) {
    		msm.mergeReplicaPolicyInfo(mergeTarget, sysMetas);
    		//    		} else if (mergeMethod.equalsIgnoreCase("rightsHolder")) {
    		//    			msm.mergeRightsHolder(mergeTarget, sysMetas);
    	} else {
    		throw new Exception("unknown mergeMethod specified: " + mergeMethod);
    	}
    	SystemMetadata actual = sysMetas.get(mergeTarget);

    	// display the expected sysmeta 
    	if (logger.isDebugEnabled()) {
    		System.out.println("Expected (sortedArrays = "+ sort + "):");
    		for (String line : TypeCompareUtil.reportSubtypesListing(expected, sort)) {
    			System.out.println(line);
    		}
    		System.out.println("--------------------------------------");
    	}

    	// display the actual sysmeta 
    	if (logger.isDebugEnabled()) {
    		System.out.println("Actual (sortedArrays = "+ sort + "):");
    		for (String line : TypeCompareUtil.reportSubtypesListing(actual,sort)) {
    			System.out.println(line);
    		}
    		System.out.println("--------------------------------------");
    	}

    	Map<String,Object> labeledSysmeta = new LinkedHashMap<String,Object>();

    	labeledSysmeta.put("actual", actual);
    	labeledSysmeta.put("expected", expected);

    	List<String> diffs = TypeCompareUtil.compareD1ServiceType(labeledSysmeta, sort);
    	for (String line : diffs) {
    		System.out.println(line);
    	}
    	Assert.assertTrue("Should not get any differences",diffs.size()<2);

    	if (level != null) {

    		if (report.getWarningList().size() != 0) {

    			boolean foundIt = false;
    			for(String warning : report.getWarningList()) {
    				if (warning.contains(messageContains)) {
    					foundIt = true;
    					break;
    				}
    			}
    			Assert.assertTrue("Logging message should contain '" 
    					+ messageContains + "'", foundIt);

    		} else {
    			Assert.fail("The test should have produced a warning.");
    		}
    	}
    	return report;

    }
   
    private static SystemMetadata clone(SystemMetadata sysmeta) {
    	return (SystemMetadata) SerializationUtils.clone(sysmeta);
    }
    	
    
    class TestAppender implements Appender {
        
    	private  List<LoggingEvent> log = new ArrayList<LoggingEvent>();
        private String name;
        private List<Filter> filters = new ArrayList<Filter>();
        ErrorHandler testEh = new org.apache.log4j.spi.ErrorHandler();
        Layout layout = new Layout();
    	@Override
    	public boolean requiresLayout() {
    		return false;
    	}

    	public void append(final LoggingEvent loggingEvent) {
    		log.add(loggingEvent);
    	}
    	@Override
    	public void close() {
    	}

    	public List<LoggingEvent> getLog() {
    		return new ArrayList<LoggingEvent>(log);
    	}

        @Override
        public void addFilter(Filter filter) {
            filters.add(filter);
        }

        @Override
        public Filter getFilter() {
            return filters.remove(0);
        }

        @Override
        public void clearFilters() {
            filters.clear();
        }

        @Override
        public void doAppend(LoggingEvent le) {
            this.append(le);
        }

        @Override
        public String getName() {
            return this.name;
        }

        @Override
        public void setErrorHandler(ErrorHandler eh) {
            testEh = eh;
        }

        @Override
        public ErrorHandler getErrorHandler() {
           return this.testEh;
        }

        @Override
        public void setLayout(Layout layout) {
            this.layout = layout;
        }

        @Override
        public Layout getLayout() {
            return this.layout;
        }

        @Override
        public void setName(String string) {
            name = string;
        }
    }	
}

