/**
 * This work was created by participants in the DataONE project, and is
 * jointly copyrighted by participating institutions in DataONE. For 
 * more information on DataONE, see our web site at http://dataone.org.
 *
 *   Copyright ${year}
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 * 
 * $Id$
 */
package org.dataone.tidy.merge.strategy;

import java.io.IOException;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.dataone.service.exceptions.NotFound;
import org.dataone.service.exceptions.ServiceFailure;
import org.dataone.service.types.TypeCompareUtil;
import org.dataone.service.types.v1.Identifier;
import org.dataone.service.types.v1.SystemMetadata;
import org.dataone.service.util.TypeMarshaller;
import org.dataone.tidy.merge.MNSystemMetadataDelegate;
import org.dataone.tidy.merge.MergeReport;
import org.jibx.runtime.JiBXException;
import org.junit.Assert;
import org.junit.Test;

public class MergeSystemMetadata_PublicMethodTest {


    protected static SystemMetadata minimalSysmeta = null;
    

    @Test 
    public void testTrue() {
    	
    }
    
    @Test
    public void testMend_replicationAllowed() throws Exception
    {
    	String testPath = "replicationAllowed/";
    	
    	Map<String, SystemMetadata> sysmetas = new LinkedHashMap<String, SystemMetadata>();

    	sysmetas.put("urn:node:cnORC",  getSysmetaResource(testPath + "sysmeta_nonAsc_1.xml"));
    	sysmetas.put("urn:node:cnUCSB", getSysmetaResource(testPath + "sysmeta_nonAsc_2.xml"));
    	sysmetas.put("urn:node:cnUNM",  getSysmetaResource(testPath + "sysmeta_asc.xml"));

    	SystemMetadata expected = getSysmetaResource(testPath + "sysmeta_mended.xml");
    	
    	mergeAndCompare(sysmetas, testPath + "sysmeta_authMN.xml", expected);
    }
    
    @Test
    public void testMend_replicationNotAllowed() throws Exception
    {
    	String testPath = "replicationNotAllowed/";
    	
    	Map<String, SystemMetadata> sysmetas = new LinkedHashMap<String, SystemMetadata>();

    	sysmetas.put("urn:node:cnORC",  getSysmetaResource(testPath + "sysmeta_nonAsc_1.xml"));
    	sysmetas.put("urn:node:cnUCSB", getSysmetaResource(testPath + "sysmeta_nonAsc_2.xml"));
    	sysmetas.put("urn:node:cnUNM",  getSysmetaResource(testPath + "sysmeta_asc.xml"));

    	SystemMetadata expected = getSysmetaResource(testPath + "sysmeta_mended.xml");
    	
    	mergeAndCompare(sysmetas, testPath + "sysmeta_authMN.xml", expected);
    }
    
    @Test
    public void testMend_ObjectFormatList_SpecialHandling() throws Exception
    {
    	String testPath = "replicationNotAllowed/";
    	
    	Map<String, SystemMetadata> sysmetas = new LinkedHashMap<String, SystemMetadata>();

    	sysmetas.put("urn:node:cnORC",  getSysmetaResource(testPath + "sysmeta_nonAsc_1.xml"));
    	sysmetas.put("urn:node:cnUCSB", getSysmetaResource(testPath + "sysmeta_nonAsc_2.xml"));
    	sysmetas.put("urn:node:cnUNM",  getSysmetaResource(testPath + "sysmeta_asc.xml"));

    	SystemMetadata expected = getSysmetaResource(testPath + "sysmeta_mended.xml");
    	
    	MergeReport report = new MergeReport(new Identifier());
    	report.getIdentifier().setValue("OBJECT_FORMAT_LIST.1.3");
    	
    	
    	MNSystemMetadataDelegate mnSysmetaGetter = new FileAuthMNSysmetaProxy(testPath + "sysmeta_authMN.xml");
    	
        MergeSystemMetadata msm = new MergeSystemMetadata(report,mnSysmetaGetter);
        report = msm.mend(sysmetas);

        List<String> diffReport = TypeCompareUtil.compareSystemMetadata(
        		report.getResultingSystemMetadata(), 
        		expected);

        for (String line : diffReport) {
            System.out.println(line);
        }

        Assert.assertTrue("Should get Access Policy differences when authmn is bypassed", diffReport.size() > 1);
    }

    
    
    private SystemMetadata getSysmetaResource(String file) 
    throws IOException, InstantiationException, IllegalAccessException, JiBXException 
    {
    	return TypeMarshaller.unmarshalTypeFromStream(SystemMetadata.class,
    			this.getClass().getResourceAsStream(file));
    }

    
    
    
    
    private void mergeAndCompare(Map<String,SystemMetadata> sysmetas, String authMNsysmeta, SystemMetadata expectedSysmeta)
            throws IOException, InstantiationException, IllegalAccessException, JiBXException, Exception 
    {
        
    	MergeReport report = new MergeReport(new Identifier());
    	report.getIdentifier().setValue("foo");
    	
    	
    	MNSystemMetadataDelegate mnSysmetaGetter = new FileAuthMNSysmetaProxy(authMNsysmeta);
    	
        MergeSystemMetadata msm = new MergeSystemMetadata(report,mnSysmetaGetter);
        report = msm.mend(sysmetas);

        List<String> diffReport = TypeCompareUtil.compareSystemMetadata(
        		report.getResultingSystemMetadata(), 
        		expectedSysmeta);

        for (String line : diffReport) {
            System.out.println(line);
        }

        Assert.assertTrue("Should not get any differences", diffReport.size() < 2);
    }
    
    
    @Test
    public void testMend_replicationNotAllowed_ServiceFailure() throws Exception
    {
    	String testPath = "replicationNotAllowed/";
    	
    	Map<String, SystemMetadata> sysmetas = new LinkedHashMap<String, SystemMetadata>();

    	sysmetas.put("urn:node:cnORC",  getSysmetaResource(testPath + "sysmeta_nonAsc_1.xml"));
    	sysmetas.put("urn:node:cnUCSB", getSysmetaResource(testPath + "sysmeta_nonAsc_2.xml"));
    	sysmetas.put("urn:node:cnUNM",  getSysmetaResource(testPath + "sysmeta_asc.xml"));

    	SystemMetadata expected = getSysmetaResource(testPath + "sysmeta_mended.xml");
    	
//    	mergeAndCompare(sysmetas, testPath + "sysmeta_authMN.xml", expected);
    	
    	MergeReport report = new MergeReport(new Identifier());
    	report.getIdentifier().setValue("foo");
    	
    	
    	MNSystemMetadataDelegate mnSysmetaGetter = new ExceptionAuthMNSysmetaProxy(ServiceFailure.class);
    	
        MergeSystemMetadata msm = new MergeSystemMetadata(report, mnSysmetaGetter);
        report = msm.mend(sysmetas);

//        List<String> diffReport = TypeCompareUtil.compareSystemMetadata(
//        		report.getResultingSystemMetadata(), 
//        		expected);
//
//        for (String line : diffReport) {
//            System.out.println(line);
//        }

//        Assert.assertTrue("Should not get any differences", diffReport.size() < 2);
        
        System.out.println(report.getProcessingException().getMessage());
        Assert.assertTrue("Should not populate mergedSysmeta field in report",
        		report.getResultingSystemMetadata() == null);
        Assert.assertTrue("should have a processing exception in the report",
        		report.getProcessingException() != null);
    }
    
    
}

